from shapely.geometry import CAP_STYLE, JOIN_STYLE, LineString

# NOTE: This is indeed, supposed to be a string
# we care about the precision and don't wish to faff about w/ floats
YELLOW = ("1", "0.952941", "0.658824")  # RGB


def content_contains_highlight(content):
    return "{} {} {} RG".format(*YELLOW).encode("utf-8") in content


# All 'Path Construction Operators'
# section 8.5.2 from https://www.adobe.com/content/dam/acom/en/devnet/pdf/PDF32000_2008.pdf
PATH_OPS = [
    "w",
    "J",
    "j",
    "M",
    "d",
    "ri",
    "i",
    "gs",
    "q",
    "Q",
    "cm",
    "m",
    "l",
    "c",
    "v",
    "y",
    "h",
    "re",
    "S",
    "s",
    "f",
    "F",
    "f*",
    "B",
    "B*",
    "b",
    "b*",
    "n",
    "W",
    "W*",
    "BT",
    "ET",
    "Tc",
    "Tw",
    "Tz",
    "TL",
    "Tf",
    "Tr",
    "Ts",
    "i",
    "Td",
    "TD",
    "Tm",
    "T*",
    "Tj",
    "TJ",
    "'",
    '"',
    "d0",
    "d1",
    "CS",
    "cs",
    "SC",
    "SCN",
    "sc",
    "scn",
    "G",
    "g",
    "RG",
    "rg",
    "K",
    "k",
    "sh",
    "BI",
    "ID",
    "EI",
    "Do",
    "MP",
    "DP",
    "BMC",
    "BDC",
    "EMC",
    "BX",
    "EX",
]


def tokenize_graphics(raw_graphics_content):
    """Convert a raw graphics content stream into a sequence of operations and their arguments.

    PDF operations are in postfix position, so the easiest way to parse them is to use a stack.
    Arguments are added onto the stack until we see a valid operation code. When we see a valid
    operation, we pop all the arguments off of the stack and continue onto the next operation.

    To be safe, all Path Construction Operations have been included in PATH_OPS. However, there
    are indeed more operations defined in the PDF specification, and no effort is made to ensure
    that the arguments to the operations are valid.
    """
    arg_stack = []
    for token in raw_graphics_content.decode("utf-8").split():
        if token in PATH_OPS:
            yield (token, arg_stack.copy())
            arg_stack.clear()
        else:
            arg_stack.append(token)


# It so happens that these numbers are the same, but we shant rely on that remaining the same!
# This also is a nice way of documenting what the numbers actually mean.
CAP_STYLES = {1: CAP_STYLE.round, 2: CAP_STYLE.flat, 3: CAP_STYLE.square}

JOIN_STYLES = {1: JOIN_STYLE.round, 2: JOIN_STYLE.mitre, 3: JOIN_STYLE.bevel}


def highlighter_lines(raw_graphics_content):
    """Render all highlighter lines into LineStrings.

    This function implements a (very) small subset of the PDF specification for drawing paths.
    Individual highlighter strokes are rendered into shapely LineString objects. Luckily for us,
    and almost certainly not due to coincidence, the semantics for defining LineString objects is
    nearly identical those of PDF paths.

    Here's an annotated example of a highlighter path pulled from a PDF:

    q                         -- push current graphics state onto the stack
    1 0.952941 0.658824 RG    -- set the stroking color to use in RGB
    12.5480766 w              -- set the line width (total width, not from centerline)
    1 J                       -- set the line cap style to 'round'
    1 j                       -- set the line join style to 'round'
    /FXE2 gs                  -- no idea what this does
    1 0 0 1 0 0 cm            -- set the current transformation matrix
    0 0 m                     -- move the cursor without drawing a line
    397.95938 614.77747 m     -- move the cursor again, overriding the previous value
    397.48959 615.17413 l     -- add a straight line segment from the current point to this point
    396.60797 615.76379 l     -- do that many, many more times
    ...                       -- and a lot more times
    51.715569 609.7724 l      -- last time
    S                         -- stroke the path
    Q                         -- pop from the graphics state stack to restore the previous value

    Key assumptions:
      - the specified color is always exactly the same (even the precision)
      - the 'RG' operation setting the color appears before all other relevant operations
      - only straight lines are used (no curves!), this isn't the case for other markers
      - the transformation matrix (if specified) is always the identity matrix
        this is verified however, so if this assumption is violated there will be an error
    """
    lines = []

    current_line = []
    width = None
    cap_style = None
    join_style = None

    for op, args in tokenize_graphics(raw_graphics_content):
        if (op, args) == ("RG", list(YELLOW)):
            current_line = [None]  # To be overrode by the last m op
        elif current_line:
            if op == "m":  # Move the draw position
                current_line[0] = (float(args[0]), float(args[1]))
            elif op == "j":
                i_join_style = int(args[0])
                join_style = JOIN_STYLES[i_join_style]
            elif op == "J":
                i_cap_style = int(args[0])
                cap_style = CAP_STYLES[i_cap_style]
            elif op == "w":
                width = float(args[0])  # Total width, will /2 later
            elif op == "l":
                current_line.append((float(args[0]), float(args[1])))
            elif op == "S":  # Finish the line
                # Make sure we got all the params we need to draw the line correctly
                assert (
                    width is not None
                ), "Expected to see a width for the stroke before stroke end."
                assert (
                    cap_style is not None
                ), "Expected to see a cap style before stroke end."
                assert (
                    join_style is not None
                ), "Expected to see a join style before stroke end."
                assert len(current_line) > 1, "Invalid line, not enough points."

                # Draw that thang
                yield LineString(current_line).buffer(
                    width / 2, cap_style=cap_style, join_style=join_style
                )

                # Reset the state
                current_line = None
                width = None
                cap_style = None
                join_style = None
            elif op == "cm":
                if args != ["1", "0", "0", "1", "0", "0"]:
                    raise NotImplementedError(
                        "Transform matrices are not implemented, but shouldn't be hard to implement"
                    )
            else:
                pass  # We don't care about other operations



from enum import Enum
from functools import partial
from itertools import chain, groupby
from operator import itemgetter

import click
import fitz
from shapely import ops
from shapely.geometry import MultiPolygon, Polygon, box

from .parsing import content_contains_highlight, highlighter_lines
import os
import textwrap
from collections import defaultdict

# TODO: Warn if highlights appear 'sloppy' mean diff from 100% area coverage on words?

WordSelectionMethod = Enum("WordSelectionMethod", "CENTROID AREA_RATIO")
DEFAULT_AREA_RATIO = 0.5  # TODO: Allow CLI control of this


def word_is_highlighted(method, highlight_poly, word_box):
    """Check if the word box is highlighted by the given highlight polygon with the provided method.

    The centroid method is more robust against vertically drifting lines, especially as the text
    boxes tend to overlap and that can lead to more area that it would appear being selected.
    """
    if method == WordSelectionMethod.CENTROID:
        return highlight_poly.contains(word_box.centroid)
    if method == WordSelectionMethod.AREA_RATIO:
        overlap = highlight_poly.intersection(word_box)
        overlap_ratio = overlap.area / word_box.area
        return overlap_ratio > DEFAULT_AREA_RATIO
    raise ValueError(f"Unknown word selection method: {method}")


HighlightType = Enum("HighlightType", "TEXTUAL SELECTION")


def classify_highlight(highlight_poly, clipping_area_threshold):
    """Determine if the highlight polygon is a textual selection, or a clipping selection.

    If the polygon has precisely one interior ring, and is of sufficient area we will consider it
    to be a clipping selection, otherwise it will be classified as textual.
    """
    if len(highlight_poly.interiors) == 1:
        # TODO: The CropBox area can vary significantly.
        #       Maybe use a ratio of the interior to the crop box area?
        if Polygon(highlight_poly.interiors[0]).area > clipping_area_threshold:
            return HighlightType.SELECTION
    return HighlightType.TEXTUAL


def classify_highlights(polys, clipping_area_threshold):
    """Group the highlight polygons based on their classification."""
    text_highlights = []
    selection_highlights = []
    for poly in polys:
        highlight_type = classify_highlight(poly, clipping_area_threshold)
        if highlight_type == HighlightType.TEXTUAL:
            text_highlights.append(poly)
        elif highlight_type == HighlightType.SELECTION:
            selection_highlights.append(poly)
    return text_highlights, selection_highlights


def coordinate_transformer(page):
    """Provide a transformer that maps PDF graphics coordinates to Fitz coordinates (and back).

    PyMuPDF uses y=0 for the top of the page, but the PDF format (and our debug plotting code) use
    y=0 for the bottom of the page. We choose to make the PDFs into fitz coordinates because
    shapely makes such transformations very simple to express.
    """

    def to_fitz_coord(x, y):
        # PyMuPDF uses y=0 for the top, but pdf uses y=0 for the bottom
        return x, page.CropBox[-1] - y

    return partial(ops.transform, to_fitz_coord)


def extract_highlight_lines(doc, page):
    """Extract the highlighter line geometries from the page.

    All the real work is getting done in the parsing code, but it is important that we apply the
    coordinate transform immediately after the lines are extracted so we never have to deal with
    different coordinate spaces.
    """
    polys = []
    for xref in page._getContents():  # pylint: disable=protected-access
        if not doc.isStream(xref):
            continue
        content = doc.xrefStream(xref)
        if content_contains_highlight(content):
            polys.extend(highlighter_lines(content))

    to_fitz_coords = coordinate_transformer(page)
    return [to_fitz_coords(poly) for poly in polys]


def normalize_polys(geometry):
    """Turn MultiPolygons into their constituent geometries, and Polygons into singleton lists.

    When we take the union of all the highlight lines, we may get 1 or n distinct polygons. This
    is used to treat both cases identically.
    """
    if geometry.geom_type == "MultiPolygon":
        return geometry.geoms
    if geometry.geom_type == "Polygon":
        return [geometry]
    raise ValueError(f"Unexpected geometry type: {geometry.geom_type}")


def merge_highlight_lines(lines):
    """Take the union of the line geometries and break them into distinct polygons."""
    union = ops.unary_union(lines)
    return normalize_polys(union)


def extract_clips(selection_polys, page, clip_zoom):
    """Get Pixmaps for each of the given selection polygons."""
    for selection_poly in selection_polys:
        # TODO: Enable masking by the interior of the poly
        selection_rect = fitz.Rect(*selection_poly.exterior.bounds)
        # TODO: Enable rendering without the yellow lines
        yield page.getPixmap(
            clip=selection_rect.round(), matrix=fitz.Matrix(clip_zoom, clip_zoom)
        )


def extract_text_highlights(text_highlights, page, word_selection_method, max_skip_len):
    """Extract the highlighted text with the provided method, allowing up to max_skip_len skips.

    If a highlight path wanders for a couple of words but then returns, it is annoying to have two
    distinct textual selections and also not know what the intermediate words are. Thus, we allow
    for the case where a couple words are skipped. It is unlikely that two intentionally distinct
    highlights are only seperated by a couple of words, and if they are it will cause much less
    harm to include more words than to exclude them.

    The algorithm to compute the selections is greatly complicated by this requirement however.
    It is certainly possible to refactor this into a single imperative loop, but after messing up
    the implementation with that approach several times I resorted to a somewhat ugly, but more
    easy to reason about implementation based on groupby.
    """
    text_highlight = MultiPolygon(text_highlights)
    # NOTE: If the skip logic proves to be finnicky with adjacent, yet distinct polys, it is
    # possible to instead look at each polygon by itself instead of merging them all into one
    # mega MultiPolygon.
    words = page.getText("words")
    word_boxes = [
        (word, box(x0, y0, x1, y1)) for x0, y0, x1, y1, word, _, _, _ in words
    ]

    is_highlighted = partial(word_is_highlighted, word_selection_method, text_highlight)

    runs = [
        (is_selected, list(map(itemgetter(0), words)))
        for is_selected, words in groupby(
            word_boxes, key=lambda t: is_highlighted(t[1])
        )
    ]

    marked_skips = [
        (is_selected or len(run) <= max_skip_len, run) for is_selected, run in runs
    ]

    merged_runs = [
        list(chain.from_iterable(map(itemgetter(1), runs_to_merge)))
        for is_selected, runs_to_merge in groupby(marked_skips, key=itemgetter(0))
        if is_selected
    ]

    return [" ".join(run) for run in merged_runs]


def extract_highlights(
    doc,
    max_skip_len,
    word_selection_method,
    clip_area_threshold,
    clip_zoom,
    debug_render,
):
    page_text_highlights = defaultdict(list)
    page_clips = defaultdict(list)
    for page in doc:
        highlight_lines = extract_highlight_lines(doc, page)
        if not highlight_lines:
            continue

        highlight_polys = merge_highlight_lines(highlight_lines)

        text_highlight_polys, selection_highlight_polys = classify_highlights(
            highlight_polys, clip_area_threshold
        )

        if debug_render:
            # pylint: disable=import-outside-toplevel
            import matplotlib.pyplot as plt

            debug_page(page, text_highlight_polys, selection_highlight_polys)
            plt.show()

        if selection_highlight_polys:
            selection_highlights = extract_clips(
                selection_highlight_polys, page, clip_zoom
            )
            page_clips[page.number + 1].extend(selection_highlights)

        if text_highlight_polys:
            text_highlights = extract_text_highlights(
                text_highlight_polys,
                page,
                WordSelectionMethod[word_selection_method],
                max_skip_len,
            )
            page_text_highlights[page.number + 1].extend(text_highlights)

    return dict(page_text_highlights), dict(page_clips)


def debug_page(page, text_highlights, selection_highlights):
    """Plot the word boxes and highlights for a particular page.

    After calling this, you must still use plt.show() to see the result! We don't call it here to
    avoid blocking on each page, although that is still possible to do if you want.
    """
    # pylint: disable=import-outside-toplevel,too-many-locals
    import matplotlib.pyplot as plt
    import descartes

    _, _, page_width, page_height = page.CropBox
    fig = plt.figure()
    ax = fig.gca()
    ax.set_xlim((0, page_width))
    ax.set_ylim((0, page_height))
    ax.set_aspect("equal")

    flip = coordinate_transformer(page)

    words = page.getText("words")
    word_boxes = [box(x0, y0, x1, y1) for x0, y0, x1, y1, word, _, _, _ in words]
    for word_box in word_boxes:
        word_patch = descartes.PolygonPatch(flip(word_box))
        ax.add_patch(word_patch)

    for text_highlight in text_highlights:
        highlight_patch = descartes.PolygonPatch(
            flip(text_highlight), fc="yellow", alpha=0.5
        )
        ax.add_patch(highlight_patch)

    for selection_highlight in selection_highlights:
        selection_patch = descartes.PolygonPatch(
            flip(selection_highlight), fc="orange", alpha=0.7
        )
        ax.add_patch(selection_patch)


@click.command()
@click.argument("filename", type=click.Path(exists=True, dir_okay=False, readable=True))
@click.option(
    "--out",
    type=click.Path(),
    default=None,
    help="Directory for output. By default the filname without the PDF extension is used.",
)
@click.option(
    "--max-skip-len",
    default=3,
    help="Number of words that can be skipped without being excluded from a highlight.",
)
@click.option(
    "--word-selection-method",
    type=click.Choice([m.name for m in WordSelectionMethod], case_sensitive=False),
    default=WordSelectionMethod.CENTROID.name,
    help="How to determine if a particular word has been highlighted.",
)
@click.option(
    "--clip-area-threshold",
    default=1000,
    help="How much area a clipping must enclose to be considered a clipping.",
)
@click.option(
    "--clip-zoom",
    default=3,
    help="How much to zoom in for the rendering of a clip. Increase for better quality.",
)
@click.option(
    "--yes",
    default=False,
    is_flag=True,
    help="WARNING: Potentially destructive! Answer yes to any interactive prompts.",
)
@click.option(
    "--debug-render",
    is_flag=True,
    default=False,
    help="Render the highlights and text boxes for each page.",
)
def main(
    filename,
    out,
    max_skip_len,
    word_selection_method,
    clip_area_threshold,
    clip_zoom,
    yes,
    debug_render,
):
    """Extract textual highlights and clippings from FILENAME."""

    file_basename, ext = os.path.splitext(filename)
    if ext == "" and out is None:
        click.echo(
            "If the pdf has no extension, the output directory must be manually specified.",
            err=True,
        )
        raise click.Abort()
    if ext != ".pdf" and not yes:
        click.confirm(
            f"Unexpected extension: {ext} - do you want to continue?", abort=True
        )
    if out is None:
        out = file_basename + "-highlights"
    if os.path.isdir(out):
        if not yes:
            click.confirm(
                f"Output directory {out} already exists, overwrite any contents?",
                abort=True,
            )
    else:
        os.mkdir(out)

    highlights_p = os.path.join(out, "highlights.txt")

    with fitz.open(filename) as doc, open(highlights_p, "w") as highlights_f:
        page_text_highlights, page_clips = extract_highlights(
            doc,
            max_skip_len,
            word_selection_method,
            clip_area_threshold,
            clip_zoom,
            debug_render,
        )

        for page_num, text_highlights in page_text_highlights.items():
            highlights_f.write(f"page {page_num}:\n")
            for highlight in text_highlights:
                block = textwrap.indent(textwrap.fill(highlight) + "\n\n", " " * 4)
                highlights_f.write(block)

        for page_num, clips in page_clips.items():
            for i, clip in enumerate(clips):
                png_path = os.path.join(out, f"p{page_num}-c{i}.png")
                clip.writePNG(png_path)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['remarkable_highlights']

package_data = \
{'': ['*']}

install_requires = \
['PyMuPDF>=1.16.17,<2.0.0', 'click>=7.1.1,<8.0.0', 'shapely>=1.7.0,<2.0.0']

extras_require = \
{'debug': ['matplotlib>=3.2.1,<4.0.0', 'descartes>=1.1.0,<2.0.0']}

entry_points = \
{'console_scripts': ['remarkable-highlights = '
                     'remarkable_highlights.extract:main']}

setup_kwargs = {
    'name': 'remarkable-highlights',
    'version': '0.1.3',
    'description': 'Extract highlights from PDFs exported from the reMarkable web interface.',
    'long_description': '# remarkable-highlights\n\nInspired and derived from https://github.com/soulisalmed/biff\n\nAutomatically extract highlights and clippings from PDFs annotated with your reMarkable.\n',
    'author': 'Ben Longo',
    'author_email': 'benlongo9807@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/benlongo/remarkable-highlights',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
